function next_prime!(a::fmpz, proved::Bool = true)
  ccall((:fmpz_nextprime,Nemo.libflint), Nothing,
         (Ref{fmpz}, Ref{fmpz}, Int),
         a, a, proved)
  return a
end

struct Primes
  a::fmpz
end

Base.iterate(ps::Primes, ::Nothing=nothing) = (deepcopy(next_prime!(ps.a)), nothing)


function next_probable_prime(z::fmpz)
  z < 0 && error("Argument must be positive")

  Tone = one(z)
  Tzero = zero(z)
  Ttwo = fmpz(2)

  if isone(z) || iszero(z)
    return Ttwo
  end

  if iseven(z)
    z += Tone
  else
    z += Ttwo
  end

  while !isprobable_prime(z)
    Nemo.addeq!(z, Ttwo)
  end

  return z
end

################################################################################
#
#  Primes iterator
#
################################################################################

struct ProbablePrimesSet{T}
  from::T
  to::T
  nocond::Bool  # If set, there is no condition on the prime.
  proof::Bool   # Does not do anything at the moment
  mod::T        # If set (i.e. > 1), only primes p % mod == a are returned
  a::T
  sv::UInt

  function ProbablePrimesSet{T}(f::T, t::T) where {T}
    z = new{T}(f, t, true, false, T(1), T(0), UInt(1))
    return z
  end

  function ProbablePrimesSet{T}(f::T, t::T, mod::T, val::T) where {T}
    sv = UInt(1)
    p = UInt(2)
    while sv < 2^30 && p < f
      if !iszero(mod % p)
        sv *= p
      end
      p = next_prime(p, false)
    end
    if mod == 1 && val == 0
      return new{T}(f, t, false, false, mod, val, sv)
    end
    if mod < 2
      error("modulus needs to be > 1")
    end
    if val < 1 || val >= mod
      error("wrong congruence condition: value needs to be > 0 and < modulus")
    end
    if !isone(gcd(mod, val))
      error("Modulus and value need to be coprime.")
    end  
    r = new{T}(f, t, false, false, mod, val, sv)
    return r
  end
end

function ProbablePrimesSet(f::T, t::T, mod::T, val::T) where {T}
  return ProbablePrimesSet{T}(f, t, mod, val)
end

function Base.iterate(A::ProbablePrimesSet{fmpz})
  if A.to != -1 && A.from > A.to
    return nothing
  end

  if A.nocond
    if !isprobable_prime(A.from)
      p = next_probable_prime(A.from)
    else
      p = A.from
    end

    if A.to != -1 && p > A.to
      return nothing
    else
      if p % A.mod != A.a
        return nothing
      else
        return p, p
      end
    end
  end

  curr = A.from 
  c = rem(curr, A.mod)
  if A.mod > 1 && c != A.a
    curr += (-c + A.a) % A.mod
  end

  c_U = curr % A.sv
  c_M = A.mod % A.sv
  UIntone = one(UInt)
  i = zero(UInt)
  while !isone(gcd(c_U + i * c_M, A.sv)) || !isprobable_prime(fmpz(curr))
    curr += A.mod
    i += UIntone
    if A.to != -1 && curr > A.to
      break
    end
  end
  
  if A.to != -1 && curr > A.to
    return nothing
  else
    return curr, curr
  end
end

function Base.iterate(A::ProbablePrimesSet{fmpz}, p)
  if A.nocond
    if p == 2
      nextp = fmpz(3)
    else
      nextp = next_probable_prime(p)
    end

    if A.to != -1 && nextp > A.to
      return nothing
    else
      return nextp, nextp
    end
  end

  if A.mod > 1
    m = A.mod
  else
    if p==2
      nextp = fmpz(3)
    end
    m = fmpz(2)
  end
  nextp = p+m
  i = zero(UInt)
  UIntone = one(UInt)
  c_U = nextp % A.sv
  c_M = m % A.sv
  while !isone(gcd(c_U + i * c_M, A.sv)) || !isprobable_prime(nextp)
    nextp = addeq!(nextp,m)
    i += UIntone
    if A.to != -1 && nextp > A.to
      break
    end
  end

  if A.to != -1 && nextp > A.to
    return nothing
  else
    return nextp, nextp
  end
end
