function sieve(int_obstr::Vector{fmpz}, qmax::fmpz, q0::fmpz=ZZ(1))
  # We use few primes to generate arithmetic progressions; this
  # keeps the required extra computations with CRT limited and avoids
  # unnecessary iterations over empty sets of primes.
  modulus_obstr_max = fdiv(qmax, 10000)
  modulus_obstr = ZZ(1)
  primes_obstr_indices = Int[]
  for ix in 1:length(int_obstr)
    n = int_obstr[ix]
    if isprime(n)
      push!(primes_obstr_indices,ix)
      modulus_obstr *= n
      if modulus_obstr >= modulus_obstr_max
        break
      end
    end
  end
  primes_obstr = int_obstr[primes_obstr_indices]
  for n in reverse!(primes_obstr_indices)
    deleteat!(int_obstr,n)
  end
  # @assert !isempty(primes_obstr)

  # We want to impose the class of jacobi_symbol(p,q), but have
  # jacobi_symbol(q,p) available. Observe that all p are of the form
  # 24*n-1 and hence congruent to 3 modulo 4. Quadratic reciprocity
  # suggests that we split primes q by their class modulo 4.
  modulus_obstr = 4*prod(primes_obstr)
  primes_obstr_ext = vcat(fmpz[4],primes_obstr)
  crt_obstr_ext = crt_env(primes_obstr_ext)

  # This might lead to significantly more primes to be checked, but it avoids
  # modulus computations during the iteration of primes.
  q0rem = mod(q0, modulus_obstr)
  if !iszero(q0rem)
    q0 -= q0rem
  end
  
  admissible_q = fmpz[]
  # We compute all primes q for which (n / q) = \epsilon for all
  # obstruction integers and primes n.
  for ϵ in [-1,1]
    for primes_classes in Iterators.product(
      [[n for n in ZZ(1):p-1 if jacobi_symbol(n,p) != ϵ]
       for p in primes_obstr]...)
      # q = 1 mod 4, hence jacobi_symbol(p,q) = jacobi_symbol(q,p)
      primes_classes_ext = vcat(fmpz[1], collect(primes_classes))
      qmod = mod(crt(primes_classes_ext,crt_obstr_ext), modulus_obstr)
      for q in ProbablePrimesSet(q0+qmod,qmax, modulus_obstr, qmod)
        if all(jacobi_symbol(n,q) != ϵ for n in int_obstr)
          push!(admissible_q, q)
        end
      end
      # q = 3 mod 4, hence jacobi_symbol(p,q) = -jacobi_symbol(q,p)
      primes_classes_ext[1] = 3
      qmod = mod(crt(primes_classes_ext,crt_obstr_ext), modulus_obstr)
      for q in ProbablePrimesSet(q0+qmod,qmax, modulus_obstr, qmod)
        if all(jacobi_symbol(n,q) != -ϵ for n in int_obstr)
          push!(admissible_q, q)
        end
      end
    end
  end

  return admissible_q
end

# function sieve(int_obstr::Vector{fmpz}, qmax::fmpz, q0::fmpz=ZZ(1))
#   # We use few primes to generate arithmetic progressions; this
#   # keeps the required extra computations with CRT limited and avoids
#   # unnecessary iterations over empty sets of primes.
#   primes_obstr = fmpz[]
#   nonprimes_obstr = fmpz[]
#   for n in int_obstr
#     push!(isprime(n) ? primes_obstr : nonprimes_obstr, n)
#   end
#   admissible_q = fmpz[]
#   q = max(ZZ(4),q0-1)
#   while next_prime!(q) < qmax
#     ϵ = jacobi_symbol(primes_obstr[1],q)
#     if all(jacobi_symbol(n,q) == ϵ || jacobi_symbol(n,q) == 0 for n in primes_obstr[2:end]) &&
#        all(jacobi_symbol(n,q) == ϵ || jacobi_symbol(n,q) == 0 for n in nonprimes_obstr)
#       push!(admissible_q, deepcopy(q))
#     end
#   end
#   return admissible_q
# end
