#include "stdlib.h"
#include "stdio.h"
#include "pthread.h"

#include "flint/nmod_poly.h"
#include "flint/thread_pool.h"
#include "flint/ulong_extras.h"
#include "nmod_poly_eta_qexp.h"


struct check_partition_ellQ_args {
  pthread_mutex_t * lock;
  FILE * output;
  long int ell;
  long int Qnext;
  long int Qstep;
  long int Qbnd;
  long int depth;
  nmod_poly_t etainv;
};

void
check_partition_ellQ(
    void * args_void
    )
{
  struct check_partition_ellQ_args * args = (struct check_partition_ellQ_args *)args_void;

  int flag;
  int jacsym;
  long int n;
  long int ell = args->ell;
  nmod_t mod_ellQ;
  fmpz_t a,mfmpz;
  long int Q, Qbnd;
  mp_limb_t b, inv24;
  mp_limb_t ell_classes[2];
  mp_limb_t Q_classes[3];

  fmpz_init(a);
  fmpz_init(mfmpz);

  while ( 1 ) {
    pthread_mutex_lock(args->lock);
    if ( args->Qnext >= args->Qbnd ) {
      pthread_mutex_unlock(args->lock);
      break;
    }
    Q = args->Qnext;
    Qbnd = Q + args->Qstep;
    args->Qnext = Qbnd;
    printf("ACCEPTING BLOCK %li\n", Q);
    pthread_mutex_unlock(args->lock);
  
    if ( Qbnd > args->Qbnd )
      Qbnd = args->Qbnd;
    
    for ( Q = n_is_prime(Q) ? Q : n_nextprime(Q,1); Q < Qbnd; Q = n_nextprime(Q,1) ) {
      if ( Q == ell ) continue;

      nmod_init(&mod_ellQ, ell*Q);
      inv24 = nmod_inv((mp_limb_t)24, mod_ellQ);

      // find the quadratic residue classes modulo ell, divisible by Q
      ell_classes[0] = (mp_limb_t)0;
      fmpz_set_si(mfmpz, ell);
      fmpz_set_si(a, -1);
      jacsym = fmpz_jacobi(a, mfmpz); 
      fmpz_set_si(a, Q);
      while ( fmpz_jacobi(a, mfmpz) == jacsym )
        fmpz_add_ui(a, a, (unsigned long)Q);
      ell_classes[1] = (mp_limb_t)fmpz_get_ui(a);

      // find the quadratic residue classes modulo Q, divisible by ell
      Q_classes[0] = (mp_limb_t)0;
      Q_classes[1] = (mp_limb_t)ell;
      fmpz_set_si(mfmpz, Q);
      fmpz_set_si(a, Q_classes[1]);
      jacsym = fmpz_jacobi(a, mfmpz);
      fmpz_add_ui(a, a, (unsigned long)ell);
      while ( fmpz_jacobi(a, mfmpz) == jacsym )
        fmpz_add_ui(a, a, (unsigned long)ell);
      Q_classes[2] = (mp_limb_t)fmpz_get_ui(a);

      for ( int ax = 0; ax < 2; ++ax ) {
        for ( int bx = 0; bx < 3; ++bx ) {
          // we need to build the class for 24 b - 1
          b = nmod_add(ell_classes[ax], Q_classes[bx], mod_ellQ);
          b = nmod_add(b, 1, mod_ellQ);
          b = nmod_mul(b, inv24, mod_ellQ);
          n = (long int)b;
          flag = 1;
          for ( int nx = 0 ; nx < args->depth; ++nx ) {
            if ( nmod_poly_get_coeff_ui(args->etainv, n) != 0 ) {
              flag = 0;
              break;
            }
            n += (long int)mod_ellQ.n;
          }
          if ( flag ) {
            pthread_mutex_lock(args->lock);
            fprintf(args->output, "%li %lu %i %i\n", Q, b, ax, bx);
            fflush(args->output);
            pthread_mutex_unlock(args->lock);
          }
        }
      }
    }
  }

  fmpz_clear(a);
  fmpz_clear(mfmpz);
}

int
main(
    int argc,
    char ** argv
    )
{
  if ( argc != 4 )
    exit(1);

  long int ell, Qbnd;
  int nmb_threads;

  ell = strtol(argv[1], NULL, 10);
  Qbnd = strtol(argv[2], NULL, 10);
  nmb_threads = (int)strtol(argv[3], NULL, 10);

  if ( ell == 5 || ell == 7 || ell == 11 ) {
    fprintf(stderr, "ell=%li\n", ell);
    exit(1);
  }

  printf("Computing for ell=%li and Qbnd=%li with %i threads.\n", ell, Qbnd, nmb_threads);

  thread_pool_t thread_pool;
  thread_pool_handle threads[nmb_threads];
  thread_pool_init(thread_pool, nmb_threads);
  if ( thread_pool_request(thread_pool, threads, nmb_threads) != nmb_threads )
    exit(1);

  struct check_partition_ellQ_args * args = (struct check_partition_ellQ_args *)malloc(sizeof(struct check_partition_ellQ_args));

  args->lock = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
  pthread_mutex_init(args->lock, NULL);

  char filename[36+2*18];
  sprintf(filename, "partition_ellQ_candidates_ell_%li_Qbnd_%li", ell, Qbnd);
  args->output = fopen(filename, "w");

  args->ell = ell;
  args->Qnext = 7;
  args->Qstep = 1000000;
  args->Qbnd = Qbnd;

  args->depth = 5;

  nmod_poly_init(args->etainv, ell);
  nmod_poly_eta_qexp(args->etainv, -1, args->depth*ell*(Qbnd-1)+1);

  {
    FILE * df = fopen("debug", "w");
    for ( long int n = 0; n < Qbnd; ++n )
      fprintf(df, "%lu\n", nmod_poly_get_coeff_ui(args->etainv, n));
    fclose(df);
  }

  for ( long int tx = 0; tx < nmb_threads; ++tx )
    thread_pool_wake(thread_pool, threads[tx], check_partition_ellQ, (void *)(args));

  for ( long int tx = 0; tx < nmb_threads; ++tx ) {
    thread_pool_wait(thread_pool, threads[tx]);
    thread_pool_give_back(thread_pool, threads[tx]);
  }
  thread_pool_clear(thread_pool);


  fprintf(args->output, "SEARCH_COMPLETED\n");

  pthread_mutex_destroy(args->lock);
  free(args->lock);
  fclose(args->output);
  nmod_poly_clear(args->etainv);
  free(args);
}
