using Distributed
@everywhere begin
  using Nemo, Hecke
  
  function kloosterman(R::Nemo.NmodRing, a::Int, b::Int, p::Int)
    divisible(a,p) && return R(-1)
    P,x = PolynomialRing(ZZ,"x")
    cyc = cyclotomic(p,x)
    P,x = PolynomialRing(R,"x")
    zeta = -coeff(first(Hecke.factor(P(cyc)))[1],0)
    return sum(zeta^Int(a*x1+b*invmod(x1,p)) for x1 in 1:p-1)
  end
  
  l0 = 17
  lmax = 1000
  Qmax = 10000
  ncnt = 10 # number of n checked in first test
end

task = @distributed for l in collect(PrimesSet(l0,lmax))
  l = ZZ(l)
  Rl = ResidueRing(ZZ,Int(l))
  for Q in PrimesSet(ZZ(17),ZZ(Qmax))
    Q == l && continue
    # We try to falsify the first consequence of lQ^2 congruences.
    # The prefactor in the middle term of the Hecke operator relation.
    prefactor = invmod(Q,l) * jacobi_symbol(-12,Q)

    # We start with case of delta = 0
    # Flag for whether congruences for l and Q have been ruled out.
    flag = false

    n0 = l*invmod(-l,ZZ(24))
    for n in n0:24*l:n0+24*l*ncnt
      jsQ = jacobi_symbol(n,Q)
      jsQ == 0 && continue
      if !iszero(Rl(number_of_partitions(divexact(n*Q^2+1,24))) -
                 prefactor * jsQ * Rl(number_of_partitions(divexact(n+1,24))))
        flag = true
        break
      end
    end

    # In case the first relation cannot rule out congruences, we employ the second one.
    if !flag
      # Flags for whether congruences for epsilon = (24 \beta - 1/Q, Q) = +1,-1
      # have been ruled out.
      flagp = false
      flagn = false

      n0 = l*invmod(-l*Q,ZZ(24))
      for n in Iterators.countfrom(Int(n0),24*Int(l))
        n = ZZ(n)
        if !divisible(number_of_partitions(divexact(n*Q+1,24)),l)
          # epsilon = +1; special computation if Q \equiv \pm 1
          if mod(Q+jacobi_symbol(-n,Q),l) == 0
            flagn = true
            if !flagp
              # This is (24*beta-1)/Q. It is any integer m with (m/l) = 0 and (m/Q) = epsilon = +1.
              betam = first(m for m in Iterators.countfrom(Int(l),Int(l))
                            if jacobi_symbol(m,Q) == 1)
              if !iszero(kloosterman(Rl,betam,Int(invmod(24^2*l^4,Q)*n),Int(Q)))
                flagp = true
              end
            end
          # epsilon = -1; special computation if Q \equiv \pm 1
          elseif mod(Q-jacobi_symbol(-n,Q),l) == 0
            flagp = true
            if !flagn
              # This is (24*beta-1)/Q. It is any integer m with (m/l) = 0 and (m/Q) = epsilon = -1.
              betam = first(m for m in Iterators.countfrom(Int(l),Int(l))
                            if jacobi_symbol(m,Q) == 1)
              if !iszero(kloosterman(Rl,betam,Int(invmod(24^2*l^4,Q)*n),Int(Q)))
                flagn = true
              end
            end
          else
            flagp = true
            flagn = true
          end
          if flagp && flagn
            break
          end
        end
      end
    end

    # We next treat the case of delta = -1.
    # Flag for whether congruences for l and Q have been ruled out.
    flag = false

    n0 = l*invmod(-l,ZZ(24))
    for n in 23:24:23+24*ncnt
      jsl = jacobi_symbol(-n,l)
      jsl == -1 || continue
      jsQ = jacobi_symbol(n,Q)
      jsQ == 0 && continue
      if !iszero(Rl(number_of_partitions(divexact(n*Q^2+1,24))) -
                 prefactor * jsQ * Rl(number_of_partitions(divexact(n+1,24))))
        flag = true
        break
      end
    end

    # In case the first relation cannot rule out congruences, we employ the second one.
    if !flag
      # Flags for whether congruences for epsilon = (24 \beta - 1/Q, Q) = +1,-1
      # have been ruled out.
      flagp = false
      flagn = false

      n0 = invmod(-Q,ZZ(24))
      for n in Iterators.countfrom(Int(n0),24)
        n = ZZ(n)
        jsl = jacobi_symbol(-n,l)
        jsl == -1 || continue
        if !divisible(number_of_partitions(divexact(n*Q+1,24)),l)
          # epsilon = +1; special computation if Q \equiv \pm 1
          if mod(Q+jacobi_symbol(-n,Q),l) == 0
            flagn = true
            if !flagp
              # This is (24*beta-1)/Q. It is any integer m with (m/l) = -1 and (m/Q) = epsilon = +1.
              betam = first(m for m in Iterators.countfrom(1)
                            if jacobi_symbol(m,l) == -1 && jacobi_symbol(m,Q) == 1)
              if !iszero(kloosterman(Rl,betam,Int(invmod(24^2*l^4,Q)*n),Int(Q)))
                flagp = true
              end
            end
          # epsilon = -1; special computation if Q \equiv \pm 1
          elseif mod(Q-jacobi_symbol(-n,Q),l) == 0
            flagp = true
            if !flagn
              # This is (24*beta-1)/Q. It is any integer m with (m/l) = -1 and (m/Q) = epsilon = -1.
              betam = first(m for m in Iterators.countfrom(1)
                            if jacobi_symbol(m,l) == -1 && jacobi_symbol(m,Q) == -1)
              if !iszero(kloosterman(Rl,betam,Int(invmod(24^2*l^4,Q)*n),Int(Q)))
                flagn = true
              end
            end
          else
            flagp = true
            flagn = true
          end
          if flagp && flagn
            break
          end
        end
      end
    end
  end
  @info "processed" l
end
